/* copyright 1994 by Harry Langenbacher - all rights reserved  - alpha code */
#include <stdio.h>
#include "p9000.h"
#include "ICD2061A.h"

/* SET THE SCREEN RESOLUTION DEPENDENT STUFF */
#define SYSCONFIG      ( 0x00004 / 4 )
#define INTERRUPT_EN   ( 0x0000C / 4 )
#define PREHRZC        ( 0x00118 / 4 )
#define HRZSR	       ( 0x0010C / 4 )
#define HRZBR	       ( 0x00110 / 4 )
#define HRZBF	       ( 0x00114 / 4 )
#define HRZT	       ( 0x00108 / 4 )
#define PREVRTC        ( 0x00130 / 4 )
#define VRTSR	       ( 0x00124 / 4 )
#define VRTBR	       ( 0x00128 / 4 )
#define VRTBF	       ( 0x0012C / 4 )
#define VRTT	       ( 0x00120 / 4 )

#define MEMCONFIG      ( 0x00184 / 4 )
#define RFPERIOD       ( 0x00188 / 4 )
#define RLMAX	       ( 0x00190 / 4 )

int Init ( int bytes_per_pixel , DisplayModePtr currentmode ,
                                volatile int * controlregbase )
 {
  unsigned int clock_ctrl_word , ActualHertz ;
  int bytesperline , diff , bestdiff ;
  int fieldadd1 , fieldadd2 , fieldadd3 , bestfit , linebytes ;
  int field1 , field2 , field3 ;
  unsigned int sysconfig ;
  unsigned int hrzt , hrzsr , hrzbr , hrzbf ;
  unsigned int vrtt , vrtsr , vrtbr , vrtbf ;

  bytesperline =
      currentmode -> HDisplay * ( ( bytes_per_pixel == 3 ) ? 4 : bytes_per_pixel ) ;

  diff = bestdiff = bestfit = 0x7FFFFFFF ; /* ( (unsigned) -1 ) >> 1  maxed out */
  /* only 6 ** 3 posibilities - try them all and find exact or best resolution */
  fieldadd3 = 0 ;
  for ( field3 = 2 ; field3 <= 7 ; field3 ++ )
   {
    fieldadd2 = 0 ;
    for ( field2 = 1 ; field2 <= 6 ; field2 ++ )
     {
      fieldadd1 = 0 ;
      for ( field1 = 0 ; field1 <= 5 ; field1 ++ )
       {
        linebytes = fieldadd1 + fieldadd2 + fieldadd3 ; /* resultant "horizontal res" */
        diff = bytesperline - linebytes ; /* how close is it to what we want */
        if ( diff == 0 )
          goto gotfields ;
        if ( diff < 0 )
          diff = - diff ; /* absolute value */
        if ( diff < bestdiff )
         {
          bestfit = linebytes ; /* this is the best fit yet */
          bestdiff = diff ; /* this is how close it is (the error) */
         }
        if ( linebytes > bytesperline ) /*it won't get any better in this loop */
          break ; /* ok - so I can't stand a totally dumb exhaustive search */
        if ( fieldadd1 == 0 )
          fieldadd1 = 32 ;
        else
          fieldadd1 = ( fieldadd1 * 2 ) ;
       }
      if ( fieldadd2 == 0 )
        fieldadd2 = 64 ;
      else
        fieldadd2 = ( fieldadd2 * 2 ) ;
     }
    if ( fieldadd3 == 0 )
      fieldadd3 = 128 ;
    else
      fieldadd3 = ( fieldadd3 * 2 ) ;
   }
  /*falls out to here if no match found */
  /* impossible HDisplay value (at least for this value of bytes_per_pixel */
  fprintf ( stderr , "can't do %d dots/line with %d bit color - could do %d dots\n" ,
              currentmode -> HDisplay , bytes_per_pixel * 3 ,
              bestfit / ( ( bytes_per_pixel == 3 ) ? 4 : bytes_per_pixel ) ) ;
  return 0 ;

  gotfields : /* NOTE - the p9000 manual figure 60 has the fields labeled
                 backwards - but if you look at values from viper.ini
                 you'll see that it's the left-most (msb) field that
                 can have values up to 7 , so that must be field 3 . It
                 doesn't work for rectangles if you mess this up, i.e. if
                 you calculate a wrong number or you use a field value
                 illegal for that field */

  if ( fieldadd3 == 0 )
    field3 = 0 ; /* correct field value if necessary - 0 value means add 0 */
  if ( fieldadd2 == 0 )
    field2 = 0 ; /* correct field value if necessary - 0 value means add 0 */

  sysconfig = 0x00003000 | ( field1 << 14 ) | ( field2 << 17 ) | ( field3 << 20 ) ;
  printf ( "field1=%d , field2=%d , field3=%d , sysconfig=0x%X\n" , field1 , field2 ,
               field3 , sysconfig ) ;
  printf ( "fieldadd1=%d , fieldadd2=%d , fieldadd3=%d\n" , fieldadd1 , fieldadd2 ,
               fieldadd3 ) ;
  printf ( "bytesperline=%d , diff=%d , linebytes=%d , bestfit=%d\n" , bytesperline ,
              diff , linebytes , bestfit ) ;

  /* sysconfig=0x587000; */ /* CHEAT FOR 800x600 */

  hrzt  = ( currentmode -> HTotal / 4 ) - 1 ;
  hrzsr = ( ( currentmode -> HSyncEnd - currentmode -> HSyncStart ) / 4 ) - 1 ;
  hrzbr = ( ( currentmode -> HTotal - currentmode -> HSyncStart ) / 4 ) - 1 ;
  hrzbf = ( 
            ( 
              currentmode -> HDisplay +
              ( currentmode -> HTotal - currentmode -> HSyncStart )
            )
             / 4
          ) - 1 ;
  vrtt  = currentmode -> VTotal ;
  vrtsr = currentmode -> VSyncEnd - currentmode -> VSyncStart ;
  vrtbr = currentmode -> VTotal - currentmode -> VSyncStart ;
  vrtbf = vrtbr + currentmode -> VDisplay ;

  controlregbase [ SYSCONFIG ]   = sysconfig ;
  controlregbase [ INTERRUPT_EN ] = 0x00000080 ; /* Interrupts Disabled , pg 20 */

  controlregbase [ PREHRZC ]     = 0 ;        /* PREHRZC = 0 pg 20 */
  controlregbase [ HRZSR ] = hrzsr ;
  controlregbase [ HRZBR ] = hrzbr ;
  controlregbase [ HRZBF ] = hrzbf ;
  controlregbase [ HRZT ]  = hrzt ;

  controlregbase [ PREVRTC ]     = 0 ;        /* PREVRTC = 0 pg 20 */
  controlregbase [ VRTSR ] = vrtsr ;
  controlregbase [ VRTBR ] = vrtbr ;
  controlregbase [ VRTBF ] = vrtbf ;
  controlregbase [ VRTT ]  = vrtt ;

  controlregbase [ MEMCONFIG ] = MY_MEMCONFIG ; /*MEMCONFIG = 1 Buf, 2 MB pg 20 */
  controlregbase [ RFPERIOD ]  = 0x00000186 ; /*RFPERIOD = 0x186 SYSCLK's pg 20 */
  controlregbase [ RLMAX ]     = 0x000000FA ; /*RLMAX = FA pg. 20 */
  controlregbase [ SRTCTL ]    = MY_SRTCTL ; /* load srtctl reg, enable video pg 20
                                                 but 1e5 from viper.ini instead of
                                                 1c4 or 1e4 from pg 20 */

  clock_ctrl_word = ICD2061ACalcClock ( 1000000 * currentmode -> Clock , 0 ) ;

  ICD2061ASetClock ( clock_ctrl_word ) ;

  clock_ctrl_word = ICD2061ACalcClock ( MY_MEMSPEED , 3 ) ;

  ICD2061ASetClock ( clock_ctrl_word ) ;

  ActualHertz = ICD2061AGetClock ( clock_ctrl_word ) ;
  fprintf ( stderr , "actual freq=%d\n" , ActualHertz ) ;

  return 1 ;
 }
